/***********************************************************************************************************************
 *
 * blueBill Core - open source birding
 * Copyright (C) 2009-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://bluebill.tidalwave.it
 * SCM: https://kenai.com/hg/bluebill~core-src
 *
 **********************************************************************************************************************/
package it.tidalwave.observation.simple;

import javax.annotation.Nonnegative;
import javax.annotation.Nonnull;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import it.tidalwave.util.Id;
import it.tidalwave.util.Initializer;
import it.tidalwave.role.spi.DefaultIdentifiable;
import it.tidalwave.observation.Cardinality;
import it.tidalwave.observation.Location;
import it.tidalwave.observation.Observable;
import it.tidalwave.observation.Observation;
import it.tidalwave.observation.ObservationItem;
import it.tidalwave.observation.Observer;
import it.tidalwave.observation.Source;

/*******************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 ******************************************************************************/
public abstract class SimpleObservationBaseBuilder<T> implements Observation.BaseBuilder<T>
  {
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    protected static class ObservableWithCardinality
      {
        @Nonnull
        public final Observable observable;

        @Nonnull
        public final Cardinality cardinality;

        private ObservableWithCardinality (final @Nonnull Observable observable,
                                           final @Nonnull Cardinality cardinality)
          {
            this.observable = observable;
            this.cardinality = cardinality;
          }

        @Override
        @Nonnull
        public String toString()
          {
            return String.format("%s x %s", cardinality, observable);
          }
      }
    
    @Nonnull
    protected final SimpleObservationSet observationSet;

    protected Date date;

    protected Location location;

    protected final List<Object> extras = new ArrayList<Object>();
    
    protected Source origin;

    protected final Set<Observer> observers = new HashSet<Observer>();

    protected final List<ObservableWithCardinality> observablesWithCardinality = new ArrayList<ObservableWithCardinality>();
    
    protected final Map<ObservableWithCardinality, Initializer<ObservationItem>> initializerMap =
            new HashMap<ObservableWithCardinality, Initializer<ObservationItem>>();

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    protected SimpleObservationBaseBuilder (final @Nonnull SimpleObservationSet observationSet)
      {
        this.observationSet = observationSet;
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Nonnull
    public T at (final @Nonnull Date date)
      {
        this.date = date;
        return (T)this;
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Nonnull
    public T at (final @Nonnull Location location)
      {
        this.location = location;
        return (T)this;
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Nonnull
    public T of (final @Nonnull Observable observable,
                   final @Nonnull Cardinality count)
      {
        return of(observable, count, null);
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Nonnull
    public T of (final @Nonnull Observable observable,
                   final @Nonnull Cardinality cardinality,
                   final @Nonnull Initializer<ObservationItem> initializer)
      {
        final ObservableWithCardinality observableWithCardinality = new ObservableWithCardinality(observable, cardinality);
        observablesWithCardinality.add(observableWithCardinality);

        if (initializer != null)
          {
            initializerMap.put(observableWithCardinality, initializer);
          }
        
        return (T)this;
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Nonnegative
    public int getItemCount()
      {
        return observablesWithCardinality.size();
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Nonnull
    public T by (final @Nonnull Observer observer)
      {
        observers.add(observer);
        return (T)this;
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Nonnull
    public T from (final @Nonnull Source origin)
      {
        this.origin = origin;
        return (T)this;
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Nonnull
    public T with (final @Nonnull Object extra)
      {
        this.extras.add((extra instanceof Id) ? new DefaultIdentifiable((Id)extra) : extra);
        return (T)this;
      }
  }
