/***********************************************************************************************************************
 *
 * blueBill Core - open source birding
 * Copyright (C) 2009-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://bluebill.tidalwave.it
 * SCM: https://kenai.com/hg/bluebill~core-src
 *
 **********************************************************************************************************************/
package it.tidalwave.observation.simple;

import javax.annotation.Nonnull;
import java.util.Iterator;
import java.util.HashSet;
import java.util.Set;
import it.tidalwave.util.Initializer;
import it.tidalwave.observation.Observation;
import it.tidalwave.observation.ObservationItem;
import it.tidalwave.observation.Observation.ChangeBuilder;
import it.tidalwave.observation.Observer;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
public class SimpleObservationChangeBuilder extends SimpleObservationBaseBuilder<ChangeBuilder>
                                            implements ChangeBuilder
  {
    @Nonnull
    private final Observation originalObservation;

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    protected SimpleObservationChangeBuilder (final @Nonnull SimpleObservationSet observationSet,
                                               final @Nonnull Observation originalObservation)
      {
        super(observationSet);
        this.originalObservation = originalObservation;
        at(originalObservation.getDate());
        at(originalObservation.getLocation());

        for (final Observer observer : originalObservation.findObservers().results())
          {
            by(observer);
          }

        for (final ObservationItem observationItem : originalObservation.findObservationItems().results())
          {
            of(observationItem.getObservable(), observationItem.getCardinality());
          }
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public Observation apply()
      {
        if (date == null)
          {
            return null;
          }

        final Set<ObservationItem> observationItems = new HashSet<ObservationItem>();
        final SimpleObservation observation = new SimpleObservation(observationSet, date, location, observers, observationItems, origin, extras);

        for (final ObservableWithCardinality observableWithCardinality : observablesWithCardinality)
          {
            ObservationItem observationItem = new SimpleObservationItem(observation,
                                                                        observableWithCardinality.observable,
                                                                        observableWithCardinality.cardinality);
            observationItems.add(observationItem);
            final Initializer<ObservationItem> initializer = initializerMap.get(observableWithCardinality);

            if (initializer != null)
              {
                initializer.initialize(observationItem);
              }
          }

        observationSet.replaceObservation(originalObservation, observation);
        return observation;
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Nonnull
    public ChangeBuilder without (final @Nonnull ObservationItem observationItem)
      {
        for (final Iterator<ObservableWithCardinality> i = observablesWithCardinality.iterator(); i.hasNext(); )
          {
            final ObservableWithCardinality observableWithCardinality = i.next();

            // TODO: is it enough to compare the cardinality? What about extra attributes?
            if (observableWithCardinality.observable.equals(observationItem.getObservable()) &&
                observableWithCardinality.cardinality.equals(observationItem.getCardinality()))
              {
                i.remove();
                initializerMap.remove(observableWithCardinality);
                break;
              }
          }

        return this;
      }
  }
